#!/usr/bin/env python3
from functools import partial

import py_cui
import _curses
import subprocess
import sys
import time
import threading
import os
import voxl_elrs_tools.crsf as crsf
import voxl_elrs_tools.elrs_device
import voxl_elrs_wizard.wizard_utils
import voxl_elrs_tools.configurator as configurator
from voxl_elrs_tools.crsf import PacketsTypes
import copy
from voxl_utils_python.voxl_utils import is_voxl
from voxl_elrs_tools.manage_firmware import ELRSFirmwareManager

class WidgetWriter:
    def __init__(self, widget = None):
        self.widget = widget

    def write(self, text) :
        if self.widget is not None:
            self.widget.write(text)

    def flush(self):
        pass

    def clear(self):
        if self.widget is not None:
            self.widget.clear()

    def set_target(self, target):
        self.widget = target


writer = WidgetWriter()

PWM_MODES = [
    "Standard"
]


def run_bind(obj_ref):
    stat = obj_ref.using_uart(timeout=1)
    time.sleep(0.5) # Make sure the popup has already appeared
    obj_ref.controller.stop_loading_popup()
    if not stat:
        obj_ref.controller.show_warning_popup("UART busy!", "Please wait a moment and try again")
        return

    dev_info = crsf.CRSFBind(obj_ref.device).run()
    obj_ref.release_uart()


def run_unbind(obj_ref):
    stat = obj_ref.using_uart(timeout=1)
    time.sleep(0.5) # Make sure the popup has already appeared
    obj_ref.controller.stop_loading_popup()
    if not stat:
        obj_ref.controller.show_warning_popup("UART busy!", "Please wait a moment and try again")
        return

    dev_info = crsf.CRSFUnbind(obj_ref.device).run()
    obj_ref.release_uart()


def run_boot(obj_ref):
    stat = obj_ref.using_uart(timeout=1)
    time.sleep(0.5) # Make sure the popup has already appeared
    obj_ref.controller.stop_loading_popup()
    if not stat:
        obj_ref.controller.show_warning_popup("UART busy!", "Please wait a moment and try again")
        return

    dev_info = crsf.CRSFSetBootloader(obj_ref.device).run()
    obj_ref.release_uart()


def run_recovery(obj_ref):
    stat = obj_ref.using_uart(timeout=1)
    time.sleep(0.5) # Make sure the popup has already appeared
    obj_ref.controller.stop_loading_popup()
    if not stat:
        obj_ref.controller.show_warning_popup("UART busy!", "Please wait a moment and try again")
        return
    print("Starting Recovery!")

    import esptool
    obj_ref.device.close()

    # Erase receiver
    rx_erase_command = [
        "--port", obj_ref.device.device if not is_voxl() else '/dev/slpi-uart-7',
        "--before", "no_reset",
        "erase_flash"
    ]
    try:
        esptool.main(rx_erase_command)
        result = 0
    except Exception as e:
        result = 1

    if result == 0:
        obj_ref.controller.show_message_popup("Recovery Successful!", "Please choose the firmware for your device to load the ELRS application", color=py_cui.GREEN_ON_BLACK)
    else:
        obj_ref.controller.show_error_popup("Recovery failed", "Status: {}".format(result))

    obj_ref.device.open_device()
    obj_ref.release_uart()


def get_device_info(obj_ref):
    while obj_ref.device_info is None:
        obj_ref.using_uart()
        dev_info = crsf.CRSFDev(obj_ref.device).send()
        obj_ref.release_uart()
        time.sleep(1)


def flash_firmware(obj_ref, firmware_str = None) -> None:
    stat = obj_ref.using_uart(timeout=1)
    time.sleep(0.5) # Make sure the popup has already appeared
    obj_ref.controller.stop_loading_popup()
    if not stat:
        obj_ref.controller.show_warning_popup("UART busy!", "Please wait a moment and try again")
        return

    obj_ref.flashing_widget_set.write_area.set_border_color(py_cui.WHITE_ON_BLACK)

    if obj_ref.flash_selected_tx:
        print("Flashing firmware {}".format(firmware_str))
        status = obj_ref.firmware_manager.flash_tx(firmware_str)
    else:
        if firmware_str is not None:
            print("Flashing firmware {}".format(firmware_str))
            status = obj_ref.firmware_manager.flash_from_string(firmware_str)
        else:
            print("Auto updating")
            status = obj_ref.firmware_manager.auto_update(obj_ref.device_info["name"], obj_ref.device_info["sw_ver"])

    if firmware_str is None and status == 1:
        obj_ref.controller.show_message_popup("Receiver Already Up To Date", "", color=py_cui.GREEN_ON_BLACK)
    elif status == 0:
        # print("Flashing successful!")
        obj_ref.controller.show_message_popup("Flashing successful!", "", color=py_cui.GREEN_ON_BLACK)
        obj_ref.flashing_widget_set.write_area.set_border_color(py_cui.GREEN_ON_BLACK)

        # Clear log on success
        if writer.clear:
            writer.clear()
    else:
        #print("Flashing failed {}".format(status))
        obj_ref.controller.show_error_popup("Flashing failed", "Status: {}".format(status))
        obj_ref.flashing_widget_set.write_area.set_border_color(py_cui.RED_ON_BLACK)

    obj_ref.release_uart()

def upload_config(obj_ref) -> None:
    stat = obj_ref.using_uart(timeout=1)
    time.sleep(0.5) # Make sure the popup has already appeared
    obj_ref.controller.stop_loading_popup()
    obj_ref.save_conf_from_ui()
    threading.Thread(target=obj_ref.configurator.upload_configuration, kwargs={"try_update": False, "wizard": obj_ref}, daemon=True).start()
    obj_ref.controller.show_loading_icon_popup("Uploading Configuration", "")

    obj_ref.release_uart()

def save_config(obj_ref) -> None:
    obj_ref.save_conf_from_ui()
    threading.Thread(target=obj_ref.configurator.save_data, kwargs={"wizard": obj_ref}, daemon=True).start()
    obj_ref.controller.show_loading_icon_popup("Saving Configuration", "")


def format_status(lq: int, rssi1: int, rssi2: int, rate_i: int, response: str) -> str:
    """
    Turn the received data from ELRS into text to display

    TODO: Also display channel data
    """
    return f"""Device Status

Link Quality: {lq}
RSSI Antenna 1: {rssi1}
RSSI Antenna 2: {rssi2}
Rate Index: {rate_i}

"""


def crsf_sever(obj_ref) -> None:
    """
    This runs in the backround, reading crsf packets and
    """
    reader = crsf.CRSFReader(obj_ref.device)
    dev_info_parser = crsf.CRSFDev(obj_ref.device)
    lq = 0
    rssi1 = 0
    rssi2 = 0
    rate_i = 0
    response_data = ''
    while True:
        time.sleep(0.05)
        if obj_ref.run_crsf_server:
            obj_ref.using_uart()
            try:
                # TODO: parse channels data too
                ptype, response = reader.get_packet(None, timeout=0.25)

                if ptype is not None:
                    if ptype == PacketsTypes.LINK_STATISTICS:
                        pkt = reader.parse_packet_data(ptype, response)
                        if pkt is not None:
                            lq = pkt.__getattribute__("up_link_quality")
                            rssi1 = -pkt.__getattribute__("up_rssi_ant1")
                            rssi2 = -pkt.__getattribute__("up_rssi_ant2")
                            rate_i = pkt.__getattribute__("rf_profile")

                    if ptype == PacketsTypes.RC_CHANNELS_PACKED:
                        response_data = response.hex()

                    if ptype == PacketsTypes.DEVICE_INFO:
                        dev_info = dev_info_parser.parse_info(response)
                        if dev_info is not None:
                            obj_ref.set_device_info(dev_info)

                    if obj_ref.status_block is not None:
                        obj_ref.status_block.set_title(format_status(lq, rssi1, rssi2, rate_i, response_data))
            except:
                if obj_ref.status_block is not None:
                    obj_ref.status_block.set_title("Device Status")
            obj_ref.release_uart()


class VoxlElrsMain:

    def __init__(self, controller : py_cui.PyCUI, start_uart: bool = True, no_scan: bool = False) -> None:
        self.controller = controller
        self.device = voxl_elrs_tools.elrs_device.ELRSDeviceManager()
        self.start_uart = start_uart

        self.firmware_manager = ELRSFirmwareManager(self.device)
        self.configurator = configurator.ELRSConfigurator(self.device)
        self.configurator.load_data()

        self.home_widget_set = None
        self.configurator_widget_set = None
        self.flashing_widget_set = None
        self.recovery_widget_set = None

        self.status_block = None

        # Style configuration
        self.button_x_pad = 1
        self.button_y_pad = 0
        self.menu_x_pad = 1
        self.menu_y_pad = 0

        # Extra data
        self.menu_item_list = []
        self.device_info = None
        self.flash_selected_tx = False
        self.run_crsf_server = True

        # Two home menu options, one minimal one for smaller terminals
        self.open_home_widget_set()

        if self.start_uart:
            self.device.open_device()
            # self.device.start_management_thread()

            # start the crsf server
            time.sleep(1)
            threading.Thread(target=crsf_sever, args=(self,), daemon=True).start()
            if not no_scan:
                threading.Thread(target=get_device_info, args=(self,), daemon=True).start()

    def using_uart(self, timeout=-1):
        return self.device.uart_lock.acquire(timeout=timeout)

    def release_uart(self):
        return self.device.uart_lock.release()

    def button_press(self, target):
        def exec_btn():
            threading.Thread(target=target, args=(self, ), daemon=True).start()
            self.controller.show_loading_icon_popup("Waiting for UART", "")
        return exec_btn

    def home_menu_nav(self, item: str) -> None:
        if item in self.menu_item_list:
            if item == "Flash Receiver":
                self.flash_selected_tx = False
                def receiver_flashing(auto: bool) -> None:
                    if auto:
                        threading.Thread(target=flash_firmware, args=(self, None), daemon=True).start()
                self.open_flashing_widget_set()
                # Can't auto-update if no devinfo
                if self.device_info is not None:
                    self.controller.show_yes_no_popup("Do you want to update automatically?", receiver_flashing)
            elif item == "Flash Transmitter":
                self.flash_selected_tx = True
                self.open_flashing_widget_set()
            elif item == "Configuration Menu":
                self.open_configuration_widget_set()
            elif item == "Recover Receiver":
                def recovery_flashing(conf: bool) -> None:
                    if conf:
                        threading.Thread(target=run_recovery, args=(self, ), daemon=True).start()
                        self.controller.show_loading_icon_popup("Waiting for UART", "")
                self.open_flashing_widget_set()
                self.controller.show_yes_no_popup("Are you sure you want to erase this receiver?", recovery_flashing)

    def create_home_widget_set(self) -> None:
        # no need to recreate the widgets
        if self.home_widget_set is not None:
            return

        # Create the widget set
        self.home_widget_set = self.controller.create_new_widget_set(8,8)

        # Command buttons
        button_bind = self.home_widget_set.add_button("Bind", 7, 0, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=self.button_press(run_bind))
        button_unbind = self.home_widget_set.add_button("Unbind", 7, 1, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=self.button_press(run_unbind))
        button_reboot = self.home_widget_set.add_button("Reboot", 7, 2, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=self.button_press(run_boot))
        button_reboot.set_color(py_cui.RED_ON_BLACK)
        button_bind.set_color(py_cui.GREEN_ON_BLACK)
        button_unbind.set_color(py_cui.GREEN_ON_BLACK)

        button_bind.add_mouse_command(py_cui.keys.LEFT_MOUSE_CLICK, self.button_press(run_bind))
        button_unbind.add_mouse_command(py_cui.keys.LEFT_MOUSE_CLICK, self.button_press(run_unbind))
        button_reboot.add_mouse_command(py_cui.keys.LEFT_MOUSE_CLICK, self.button_press(run_boot))

        # Menu selection
        self.menu_item_list = ["Flash Receiver", "Flash Transmitter", "Configuration Menu", "Recover Receiver"]
        menu = self.home_widget_set.add_scroll_menu("Menu", 1, 0, 6, 3, padx=self.button_x_pad, pady=self.button_y_pad)
        menu.add_item_list(self.menu_item_list)
        def home_menu_callback():
            self.home_menu_nav(menu.get())
        menu.add_key_command(py_cui.keys.KEY_ENTER, home_menu_callback)

        # Device info
        dev_info_box = self.home_widget_set.add_block_label("Device info\nLoading...", 0, 0, row_span=1, column_span=3, padx=self.menu_x_pad, pady=self.menu_y_pad)
        dev_info_box.toggle_border()
        self.home_widget_set.dev_info_box = dev_info_box

        # Device status
        status_block = self.home_widget_set.add_block_label("Device Status", 0, 3, row_span=5, column_span=5, padx=self.menu_x_pad, pady=self.menu_y_pad)
        #status_block.toggle_border()
        self.status_block = status_block

        ## DEBUG
        log_block = self.home_widget_set.add_text_block("Log", 5, 3, 3, 5, padx=self.menu_x_pad, pady=self.menu_y_pad)
        #status_block.toggle_border()
        self.home_widget_set.log_block = log_block

    def create_min_home_widget_set(self) -> None:
        # this could be used to replace an already-made version if the screen size changed
        if self.home_widget_set is not None:
            del self.home_widget_set

        # Create the widget set
        self.home_widget_set = self.controller.create_new_widget_set(2,4)

        # Command buttons
        button_bind = self.home_widget_set.add_button("Bind", 0, 3, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=lambda : (threading.Thread(target=run_bind, args=(self,)).start()))
        button_unbind = self.home_widget_set.add_button("Unbind", 1, 3, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=lambda : (threading.Thread(target=run_unbind, args=(self,)).start()))
        button_bind.set_color(py_cui.GREEN_ON_BLACK)
        button_unbind.set_color(py_cui.GREEN_ON_BLACK)

        # Menu selection
        self.menu_item_list = ["Flash Receiver", "Flash Transmitter", "Configuration Menu"]
        menu = self.home_widget_set.add_scroll_menu("Menu", 0, 0, 2, 3, padx=self.button_x_pad, pady=self.button_y_pad)
        menu.add_item_list(self.menu_item_list)
        def home_menu_callback():
            self.home_menu_nav(menu.get())
        menu.add_key_command(py_cui.keys.KEY_ENTER, home_menu_callback)

        # Device info
        # dev_info_box = self.home_widget_set.add_block_label("Device info\nLoading...", 0, 0, row_span=1, column_span=3, padx=self.menu_x_pad, pady=self.menu_y_pad)
        # dev_info_box.toggle_border()
        self.home_widget_set.dev_info_box = None

    def create_flashing_widget_set(self) -> None:
        if self.flashing_widget_set is not None:
            return

        self.flashing_widget_set = self.controller.create_new_widget_set(3,3)

        def flash_firmware_callback():
            threading.Thread(target=flash_firmware, args=(self, self.flashing_widget_set.fw_menu.get()), daemon=True).start()
            self.controller.show_loading_icon_popup("Waiting for UART", "")

        # Firmware menu (top left)
        fw_menu = self.flashing_widget_set.add_scroll_menu("Choose Firmware", 0, 0, 2, 1, padx=self.menu_x_pad, pady=self.menu_y_pad)
        self.flashing_widget_set.fw_menu = fw_menu
        fw_menu.add_key_command(py_cui.keys.KEY_ENTER, flash_firmware_callback)

        # Flash button (bottom left)
        flash_button = self.flashing_widget_set.add_button("Flash", 2, 0, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=flash_firmware_callback)

        # Flashing progress (right side) TODO: build non-editable version
        write_area = self.flashing_widget_set.add_text_block("Flashing status log", 0, 1, 3, 2, padx=self.menu_x_pad, pady=self.menu_y_pad)
        self.flashing_widget_set.write_area = write_area

        self.flashing_widget_set.add_key_command(py_cui.keys.KEY_H_LOWER, self.open_home_widget_set)

    def load_all_pwm_checks(self) -> None:
        """
        Used when first loading the UI to set which PWMs are enabled
        """
        conf = self.configurator.get_configuration()
        pwm_list = conf["pwm"]
        for index, pwm in enumerate(pwm_list):
            self.configurator_widget_set.check_menu.mark_item_index(index, pwm.get("enabled"))

    def load_pwm_from_conf(self, index: int) -> None:
        """
        Sets the PWM configuration to use the included PWM
        """
        conf = self.configurator.get_configuration()
        pwm = conf["pwm"][index]
        if pwm is not None:
            self.configurator_widget_set.mode_input.set_text(PWM_MODES[int(pwm["mode"])])
            self.configurator_widget_set.input_ch_slider.set_slider_value(int(pwm["ch"]))
            self.configurator_widget_set.failsafe_slider.set_slider_value(int(pwm["failsafe"]))
            self.configurator_widget_set.check_menu.mark_item_index(index, pwm["enabled"])
        else:
            # use default values
            self.configurator_widget_set.mode_input.set_text(PWM_MODES[0])
            self.configurator_widget_set.input_ch_slider.set_slider_value(1)
            self.configurator_widget_set.failsafe_slider.set_slider_value(-100)
            self.configurator_widget_set.check_menu.mark_item_index(index, False)

        self.configurator_widget_set.check_menu.set_selected_item_index(index)
        self.configurator_widget_set.saved_index = index

    def load_ui_from_conf(self) -> None:
        """
        This method loads the UI values from the configuration object
        """
        conf = self.configurator.get_configuration()
        self.configurator_widget_set.top_label.set_title(
            f"PWM Configuration Menu\nSelect pins from the menu on the left to change their properties. Press enter to enable/disable the pins.\nAt the moment the only pin setting taken into account is the input channel\nDo not upload configuration while connected to transmitter.\nConfiguration Name: {conf['name']}")

        if conf["model_name"]:
            self.configurator_widget_set.receiver_str.set_text(conf["model_name"])
        else:
            self.configurator_widget_set.receiver_str.set_text("")
        if conf["version"]:
            self.configurator_widget_set.version_str.set_text(conf["version"])
        else:
            self.configurator_widget_set.version_str.set_text("")
        if conf["binding_phrase"]:
            self.configurator_widget_set.binding_str.set_text(conf["binding_phrase"])
        else:
            self.configurator_widget_set.binding_str.set_text("")

        self.load_all_pwm_checks()
        self.load_pwm_from_conf(0)

    def save_pwm_index(self, conf, index) -> None:
        conf["pwm"][index]["mode"] = PWM_MODES.index(self.configurator_widget_set.mode_input.get())
        conf["pwm"][index]["ch"] = int(self.configurator_widget_set.input_ch_slider.get_slider_value())
        conf["pwm"][index]["failsafe"] = int(self.configurator_widget_set.failsafe_slider.get_slider_value())
        conf["pwm"][index]["enabled"] = self.configurator_widget_set.check_menu.get_item_index_checked(index)

    def save_conf_from_ui(self) -> None:
        """
        This method saves the state of the UI to the configuration object. It does not save it to disk
        """
        conf = self.configurator.get_configuration()
        conf["model_name"] = self.configurator_widget_set.receiver_str.get()
        conf["version"] = self.configurator_widget_set.version_str.get()
        conf["binding_phrase"] = self.configurator_widget_set.binding_str.get()

        # Save the currently selected index only since all the others should have already been saved
        if self.configurator_widget_set.saved_index is not None:
            self.save_pwm_index(conf, self.configurator_widget_set.saved_index)

        self.configurator.set_configuration(conf)

    def unimplemented_warning(self):
        self.controller.show_warning_popup("Not yet implemented!", "This feature will appear in a future voxl-elrs.")

    def create_configuration_widget_set(self) -> None:
        if self.configurator_widget_set is not None:
            return

        self.configurator_widget_set = self.controller.create_new_widget_set(8,3)

        ## Top Label
        self.configurator_widget_set.top_label = self.configurator_widget_set.add_block_label("PWM Configuration Menu", 0, 0, 1, 3, padx=self.menu_x_pad, pady=self.menu_y_pad)

        ## PWM Config
        self.configurator_widget_set.saved_index = 0
        self.configurator_widget_set.check_menu = self.configurator_widget_set.add_checkbox_menu("PWM Pins", 1, 0, 5, 1, padx=self.menu_x_pad, pady=self.menu_y_pad)
        self.configurator_widget_set.check_menu.add_item_list(["PWM Pin 0", "PWM Pin 1", "PWM Pin 2", "PWM Pin 3"])
        def pwm_selection_change(item):
            # Save here before saved_index gets updated
            self.save_conf_from_ui()
            self.load_pwm_from_conf(self.configurator_widget_set.check_menu.get_selected_item_index())
        self.configurator_widget_set.check_menu.set_on_selection_change_event(pwm_selection_change)

        # PWM Settings
        self.configurator_widget_set.mode_input = self.configurator_widget_set.add_text_box("PWM Mode", 1, 1, 1, 1, padx=self.menu_x_pad, pady=self.menu_y_pad)

        self.configurator_widget_set.input_ch_slider = self.configurator_widget_set.add_slider("Input Channel", 2, 1, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, min_val=1, max_val=16, init_val=1, step=1)
        self.configurator_widget_set.input_ch_slider.set_bar_char(' ')

        self.configurator_widget_set.failsafe_slider = self.configurator_widget_set.add_slider("Failsafe Value", 3, 1, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, min_val=-100, max_val=100, init_val=-100, step=2)
        self.configurator_widget_set.failsafe_slider.set_bar_char(' ')

        ## Other config
        self.configurator_widget_set.receiver_str = self.configurator_widget_set.add_text_box("Receiver Type (Leave empty for any)", 1, 2, 1, 1, padx=self.menu_x_pad, pady=self.menu_y_pad)
        self.configurator_widget_set.version_str = self.configurator_widget_set.add_text_box("Version Specifier (Empty will use latest)", 2, 2, 1, 1, padx=self.menu_x_pad, pady=self.menu_y_pad)
        self.configurator_widget_set.binding_str = self.configurator_widget_set.add_text_box("Binding Phrase (Leave empty for traditional binding)", 3, 2, 1, 1, padx=self.menu_x_pad, pady=self.menu_y_pad)

        ## Buttons
        self.configurator_widget_set.upload_button = self.configurator_widget_set.add_button("Upload Configuration", 6, 0, 2, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=self.button_press(upload_config))
        self.configurator_widget_set.save_button = self.configurator_widget_set.add_button("Save Configuration", 6, 1, 2, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=partial(save_config, self))
        self.configurator_widget_set.export_button = self.configurator_widget_set.add_button("Export Configuration", 6, 2, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=self.unimplemented_warning)
        self.configurator_widget_set.load_button = self.configurator_widget_set.add_button("Load Configuration", 7, 2, 1, 1, padx=self.button_x_pad, pady=self.button_y_pad, command=self.unimplemented_warning)

        self.configurator_widget_set.add_key_command(py_cui.keys.KEY_H_LOWER, self.open_home_widget_set)


    def open_home_widget_set(self) -> None:
        self.controller.set_title('voxl-elrs')
        if self.home_widget_set is None:
            self.create_home_widget_set()
        try:
            self.controller.apply_widget_set(self.home_widget_set)
            if self.device_info is not None:
                self.home_widget_set.dev_info_box.set_title(
                    "Device info\n{}\nv{}".format(self.device_info["name"], self.device_info["sw_ver"]))
            writer.set_target(self.home_widget_set.log_block)
        except py_cui.errors.PyCUIOutOfBoundsError:
            self.create_min_home_widget_set()
            self.controller.apply_widget_set(self.home_widget_set)

    def open_flashing_widget_set(self) -> None:
        self.controller.set_title('Flash Firmware')
        if self.flashing_widget_set is None:
            self.create_flashing_widget_set()

        writer.set_target(self.flashing_widget_set.write_area)
        self.flashing_widget_set.fw_menu.clear()
        # if device is known only list firmware for that device
        if self.flash_selected_tx:
            self.flashing_widget_set.fw_menu.add_item_list(self.firmware_manager.get_tx_options())
        else:
            if self.device_info is not None:
                self.flashing_widget_set.fw_menu.add_item_list(self.firmware_manager.get_options(self.device_info["name"]))
            else:
                self.flashing_widget_set.fw_menu.add_item_list(self.firmware_manager.get_options())
        self.controller.apply_widget_set(self.flashing_widget_set)

    def open_configuration_widget_set(self) -> None:
        self.controller.set_title('ELRS Configuration')
        if self.configurator_widget_set is None:
            self.create_configuration_widget_set()

        self.load_ui_from_conf()
        self.controller.apply_widget_set(self.configurator_widget_set)

    def set_device_info(self, dev_info: dict) -> None:
        self.device_info = dev_info
        if self.home_widget_set is not None:
            if self.device_info is None:
                self.home_widget_set.dev_info_box.set_title(
                    "\nFailed to determine receiver type")
            else:
                if self.home_widget_set.dev_info_box is not None:
                    self.home_widget_set.dev_info_box.set_title(
                        "Device info\n{}\nv{}".format(dev_info["name"], dev_info["sw_ver"]))


def run_wizard(no_scan: bool = False):
    voxl_elrs_wizard.wizard_utils.is_wizard = True
    root = py_cui.PyCUI(8, 8)
    voxl_elrs_wizard.wizard_utils.wizard_root = root
    # override sys.stdout here
    sys.stdout = writer

    if not is_voxl():
        root.toggle_unicode_borders()
    else:
        # root.set_border_characters(...)
        pass
    root.set_title('voxl-elrs')
    root.set_refresh_timeout(1)
    s = VoxlElrsMain(root, no_scan=no_scan)
    try:
        root.start()
    except _curses.error:
        pass

    # clear screen on end
    subprocess.run("clear")
