import ctypes as ct
import numpy as np
from time import sleep, time
import subprocess
import pdb

from .voxl_utils import is_voxl

libvoxl_io = ct.CDLL("libvoxl_io.so")

voxl_uart_init  = libvoxl_io.voxl_uart_init_legacy_slpi
voxl_uart_init.argtypes = [ct.c_int,ct.c_int]
voxl_uart_init.restype  = ct.c_int

# close an initialized serial port
voxl_uart_close = libvoxl_io.voxl_uart_close
voxl_uart_close.argtypes = [ct.c_int]
voxl_uart_close.restype  = ct.c_int

#voxl_uart_drain = libvoxl_io.voxl_uart_drain

# write bytes to an initialized serial port
voxl_uart_write          = libvoxl_io.voxl_uart_write
voxl_uart_write.argtypes = [ct.c_int, ct.POINTER(ct.c_uint8), ct.c_size_t]
voxl_uart_write.restype  = ct.c_int

# read bytes from an initialized serial port
voxl_uart_read           = libvoxl_io.voxl_uart_read_bytes
voxl_uart_read.argtypes  = [ct.c_int, ct.POINTER(ct.c_uint8), ct.c_size_t]
voxl_uart_read.restype   = ct.c_int

# flush the serial port
voxl_uart_flush = libvoxl_io.voxl_uart_flush
voxl_uart_flush.argtypes = [ct.c_int]
voxl_uart_flush.restype  = ct.c_int

voxl_uart_bytes_available = libvoxl_io.voxl_uart_bytes_available

SLPI_QUP7_UART = 17


class VoxlSerialPort():
    initialized   = False
    _baudrate      = 0
    write_timeout = 0
    port          = ''
    port_num      = SLPI_QUP7_UART
    parity        = 0
    bytesize      = 8
    stopbits      = 1
    timeout       = 0
    writeTimeout  = 0
    read_buf_ptr  = np.zeros(1024, dtype='uint8')
    read_buf_size = 256
    dtr = None
    rts = None

    def __init__(self):
        self.initialized = False

    def isOpen(self):
        return self.initialized

    def inWaiting(self):
        return self.read_buf_size

    def open(self):
        # self.port_num = int(self.port.split('-')[-1])
        self._baudrate = int(self._baudrate)
        voxl_uart_init(self.port_num, self._baudrate)
        self.initialized  = True
        # print("Opening")

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self,value):
        self._baudrate = value
        if self.initialized:
            voxl_uart_close(self.port_num)
            voxl_uart_init(self.port_num, self._baudrate)
            print('Updated baud rate to %d' % self._baudrate)

    def close(self):
        if not self.initialized:
            raise Exception('port is not initialized')

        # print("Closing")
        voxl_uart_close(self.port_num)
        # if is_voxl():
        #     subprocess.run("voxl-reset-slpi", stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        self.initialized = False

    def read(self, num_bytes_to_read, timeout=0):
        if num_bytes_to_read > self.read_buf_size:
            num_bytes_to_read = self.read_buf_size
        start_time = 0
        if timeout > 0:
            start_time = time()
        bytes_read = 0
        data_array = None
        while bytes_read == 0 and (timeout == 0 or time() < (start_time + timeout)):
            # print("READING " + str(num_bytes_to_read))
            bytes_read = voxl_uart_read(self.port_num, self.read_buf_ptr.ctypes.data_as(ct.POINTER(ct.c_uint8)), num_bytes_to_read)
            # print("GOT " + str(bytes_read))
            data_array = bytearray(self.read_buf_ptr[0:bytes_read])
            # If no timeout set, break after one try
            if timeout == 0:
                break
        return data_array

    def write(self, data):
        if not self.initialized:
            raise Exception('port is not initialized')
        data = np.array(data)
        data_ptr = data.ctypes.data_as(ct.POINTER(ct.c_uint8))
        voxl_uart_write(self.port_num,data_ptr,data.nbytes)


    def flush(self):
        voxl_uart_flush(self.port_num)

    def flushInput(self):
        pass

    def flushOutput(self):
        self.flush()

    def reset_input_buffer(self):
        pass

    def setRTS(self, *args):
        """Control flow not currently supported on voxl uart"""
        pass

    def setDTR(self, *args):
        """Control flow not currently supported on voxl uart"""
        pass

    def in_waiting(self):
        return self.inWaiting()
    
    def __del__(self):
        if self.initialized:
            self.close()

